#![no_std]
#![no_main]
#![feature(asm)]
#![feature(naked_functions)]

use core::fmt::Write;

use panic_abort as _;
use r0::zero_bss;

mod regs;
mod cortex_a9;
mod slcr;
mod uart;
use uart::Uart;
mod eth;

extern "C" {
    static mut __bss_start: u32;
    static mut __bss_end: u32;
    static mut __end: u32;
}

#[link_section = ".text.boot"]
#[no_mangle]
#[naked]
pub unsafe extern "C" fn _boot_cores() -> ! {
    use cortex_a9::{asm, regs::*};

    const CORE_MASK: u32 = 0x3;
    let stack_start = __end + 4096;

    match MPIDR.get() & CORE_MASK {
        0 => {
            SP.set(stack_start);
            boot_core0();
        }
        _ => loop {
            // if not core0, infinitely wait for events
            asm::wfe();
        },
    }
}

unsafe fn boot_core0() -> ! {
    zero_bss(&mut __bss_start, &mut __bss_end);
    main();
    panic!("return from main");
}

fn main() {
    let mut uart = Uart::uart1(115_200);
    loop {
        for i in 0.. {
            writeln!(uart, "i={}\r", i);
        }
    }

    let eth = eth::Eth::gem0();
    loop {
    }
}
